#include "pvo-utils.hh"
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include <glob.h>
#include <iostream>
#include <sstream>
#include <iomanip>

namespace PVO
{

  std::string getPSRFITSdefn(){ 
    return getenv("PSRFITSDEFN"); 
  }

  void PVOError(std::string subroutine, std::string error)
  {
    /**
     * Prints an ERROR message to the standard error stream, quoting
     * the subroutine that the problem occurred in, a descriptive
     * warning text, and sounding the bell.
     *
     * Format: ERROR \<function\> : Something bad happened.
     */
    std::string preamble = "\aERROR <" + subroutine + "> : ";
    std::cerr << preamble;
    for(unsigned int i=0;i<error.size();i++){
      std::cerr << error[i];
      if((i!=error.size()-1)&&(error[i]=='\n'))
	std::cerr << std::setw(preamble.size()-1) <<": ";
    }
  }
  //------------------------------------------------------------

  std::vector<std::string> getAllFitsFiles()
  {
    std::vector<std::string> namelist;
    glob_t list1,list2,list3;
    unsigned int i;

    list1.gl_offs = 0;
    list2.gl_offs = 0;
    list3.gl_offs = 0;

    glob("*.fits", GLOB_DOOFFS, NULL, &list1);
    glob("*.rf", GLOB_DOOFFS, NULL, &list2);
    glob("*.sf", GLOB_DOOFFS, NULL, &list3);

    namelist.resize(list1.gl_pathc+list2.gl_pathc+list3.gl_pathc);
    for(i=0;i<list1.gl_pathc+list1.gl_offs;i++) 
      namelist[i] = list1.gl_pathv[i];
    for(i=0;i<list2.gl_pathc+list2.gl_offs;i++) 
      namelist[list1.gl_pathc+i] = list2.gl_pathv[i];
    for(i=0;i<list3.gl_pathc+list3.gl_offs;i++) 
      namelist[list1.gl_pathc+list2.gl_pathc+i] = list3.gl_pathv[i];
    
    return namelist;
  }

  //------------------------------------------------------------

  std::string removeLeadingBlanks(std::string s)
  {
    /**
     * All blank spaces from the start of the string to the first
     * non-blank-space character are deleted.
     */ 
    int i=0;
    while(s[i]==' '){
      i++;
    }
    std::string newstring;
    for(unsigned int j=i;j<s.size();j++) newstring += s[j];
    return newstring;
  }
  //------------------------------------------------------------------------

  std::string deblank(std::string s)
  {
    /**
     * All blank spaces from the start of the string to the first
     * non-blank-space character, and from the last non-blank-space
     * character to the end are deleted.
     */ 
    int beg=0;
    while(s[beg]==' '){
      beg++;
    }
    int end=s.size()-1;
    while(s[end]==' '){
      end--;
    }
    std::string newstring;
    for(int j=beg;j<=end;j++) newstring += s[j];
    return newstring;
  }
  //------------------------------------------------------------------------

  std::string charToBinary(char c)
  {
    /**
     * A single byte, represented by a char type, is read in and converted
     * to a binary representation.
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param c A single byte, of char type.
     * \return A std:string that shows the binary representation of that byte.
     */

    std::stringstream ss;
    int num=c;
    if (num<0) num+=256;
    int factor = 128;
    for(int i=0;i<8;i++){
      ss << num/factor;
      num -= num/factor * factor;
      factor /= 2;
    }
    return ss.str();
  }
  //------------------------------------------------------------------------

  std::string binaryToString(std::vector<bool> v)
  {
    /**
     * A string of bits, read in as a vector of bool types, is coverted to the
     *  equivalent string of bytes, in the form of chars (in fact, a std::string).
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param v String of bits.
     * \return Std::string of bytes (each byte a char).
     */
    int size = v.size();
    int nbytes = (size-1)/8 + 1;
    std::string output;
    for(int byte=0;byte<nbytes;byte++){
      int c=0;
      for(int bit=0;bit<8;bit++){
	int pt = byte*8+bit;
	if(v[pt]) c += int(pow(2,7-bit));
      }
      //    std::cerr << int(c) << std::endl;
      output += char(c);
    }
    return output;
  }
  //------------------------------------------------------------------------

  unsigned int * binaryToInt(std::vector<bool> v)
  {
    /**
     * A string of bits, read in as a vector of bool types, is coverted to a 
     *  series of bytes, which are expressed as an array of integers (with the
     *  decimal value of each byte).
     *
     * Used for debugging the single-bit writing and reading tasks.
     *
     * \param v String of bits.
     * \return Sequence of integers, being the decimal value of each byte.
     */
    int size = v.size();
    int nbytes = (size-1)/8 + 1;
    unsigned int *output = new unsigned int[nbytes];
    for(int byte=0;byte<nbytes;byte++){
      int c=0;
      for(int bit=0;bit<8;bit++){
	int pt = byte*8+bit;
	if(v[pt]) c += (unsigned int)pow(2,7-bit);
      }
      //    std::cerr << c << std::endl;
      output[byte] = c;
    }
    return output;
  }
  //------------------------------------------------------------------------

  double dmsToDec(std::string input)
  {
    /** 
     *   Assumes the angle given is in degrees, so if passing RA as
     *   the argument, need to multiply by 15 to get the result in
     *   degrees rather than hours.  The sign of the angle is
     *   preserved, if present.
     */

    std::string dms = removeLeadingBlanks(input);

    bool isNeg = false;
    if(dms[0]=='-') isNeg = true;

    for(unsigned int i=0;i<dms.size();i++) if(dms[i]==':') dms[i]=' ';

    std::stringstream ss;
    ss.str(dms);

    double d,m,s;
    ss >> d >> m >> s;

    double dec = fabs(d) + m/60. + s/3600.;
    if(isNeg) dec = dec * -1.;

    return dec;

  }
  //------------------------------------------------------------------------

  std::string decToDMS(double input, int secondPrecision, std::string separator)
  {
    /**
     *   This is the general form, where one can specify the degree of
     *   precision of the seconds, and the separating character.
     *  \param input Angle in decimal degrees.
     *  \param secondPrecision How many decimal places to quote the seconds to.
     *  \param separator The character (or string) to use as a
     *  separator between hh and mm, and mm and ss.sss.
     */ 
    double hour = trunc(input);
    double fraction = fabs(input - hour);
    double min = trunc( fraction * 60. );
    fraction = fraction * 60. - min;
    double sec = fraction * 60.;
  
    int secondWidth = 2;
    if(secondPrecision>0) secondWidth += 1 + secondPrecision;

    std::stringstream output;
    output.setf(std::ios::fixed);
    output << std::setw(2) << std::setfill('0') << std::setprecision(0)
	   << hour << separator 
	   << std::setw(2) << std::setfill('0') << std::setprecision(0)
	   << min  << separator ;
    output << std::setw(secondWidth) << std::setfill('0') 
	   << std::setprecision(secondPrecision) << sec;

    return output.str();
  }
  //------------------------------------------------------------------------

  std::string getIAUName(double ra, double dec)
  {
    /**
     *  The positions are assumed to be J2000.0, since the name is
     *  prefixed by a J.
     *
     *  \param ra RA position, in degrees.
     *  \param dec Dec position, in degrees.
     *  \return Name of the form J1234m4321 or J1234p0123
     */

    double raHrs = ra / 15.;
    int h = int(raHrs);
    int m = (int)(fmod(raHrs,1.)*60.);
    int s = (int)(fmod(raHrs,1./60.)*3600.);
    std::stringstream ss(std::stringstream::out);
    ss.setf(std::ios::showpoint);
    ss.setf(std::ios::fixed);
    ss << "J"; 
    ss<<std::setw(2)<<std::setfill('0')<<h;
    ss<<std::setw(2)<<std::setfill('0')<<m;
    ss<<std::setw(2)<<std::setfill('0')<<s;
    int sign = int( dec / fabs(dec) );
    double d = dec / sign;
    h = int(d);
    m = (int)(fmod(d,1.)*60.);
    s = (int)(fmod(d,1./60.)*3600.);
    if(sign==1) ss<<"p"; else ss<<"m";
    ss<<std::setw(2)<<std::setfill('0')<<h;
    ss.unsetf(std::ios::showpos);
    ss<<std::setw(2)<<std::setfill('0')<<m;
    ss<<std::setw(2)<<std::setfill('0')<<s;
    return ss.str();
  }
  //------------------------------------------------------------------------

  time_t my_timegm (struct tm *tm) {
    /**
     * Local function to get the proper GM time.
     */
    time_t ret;
    char *tz;

    tz = getenv("TZ");
    setenv("TZ", "", 1);
    tzset();
    ret = mktime(tm);
    if (tz) setenv("TZ", tz, 1);
    else    unsetenv("TZ");
    tzset();
    return ret;
  }

  std::string addDay(std::string date)
  {
    /**
     *   This assumes the format is yy/mm/dd or yymmdd or yy-mm-dd
     *    -- the crucial thing is that the day number is in the last spot.
     *   The format of the date is preserved, as long as it is one of those 3.
     *  \param date Input date.
     *  \return Date of same format as input, but one day later.
     */ 

    struct tm time;
    time.tm_hour=0;
    time.tm_min =0;
    time.tm_sec =1;
    char buf[255];
    std::string output;
    time_t temp;
    if(date.size()==6) {
      strptime(date.c_str(), "%y%m%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y%m%d", gmtime(&temp));
    }
    else if(date[2]=='/'){
      strptime(date.c_str(), "%y/%m/%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y/%m/%d", gmtime(&temp));
    }
    else if(date[2]=='-'){
      strptime(date.c_str(), "%y-%m-%d", &time);
      time.tm_mday += 1;
      temp = my_timegm(&time);
      strftime(buf, sizeof(buf), "%y-%m-%d", gmtime(&temp));
    }
    else{
      std::cerr << "<addDay> : date (" << date << ") has unknown format!\n";
      return date;
    }
    output = buf;
    return output;
  
  }
  //------------------------------------------------------------------------

  std::string makeDateObs(std::string date, std::string time)
  {

    /**
     * This produces a string of the form YY-MM-DDThh:mm:ss.sss, using
     * strings containing the date and the time. It is simply a matter
     * of concatenating them, but we check the date string to make sure
     * it has the hyphens in it.
     *
     * \param date A date string, assumed to be of the format YY-MM-DD,
     * YY/MM/DD, or YYMMDD.
     * \param time A time string, assumed to be in the format
     * hh:mm:ss[.sss] (the fractional seconds are optional).
     * \return A properly formatted date/time string.
     */

    std::string date_obs;
    bool hashyphens = false;
    for(unsigned int i=0;i<date.size();i++) 
      hashyphens = hashyphens || (date[i]=='-') || (date[i] == '/');
    if(hashyphens) date_obs = date;
    else date_obs = date.substr(0,2) + "-" + 
	   date.substr(2,2) + "-" + date.substr(4,2);
    if(date[0]=='0') date_obs = "20" + date_obs;
    else             date_obs = "19" + date_obs;
    for(unsigned int i=0;i<date_obs.size();i++) 
      if(date_obs[i]=='/') date_obs[i]='-';

    date_obs += "T" + time;

    return date_obs;

  }
  //------------------------------------------------------------------------

  double angSep(std::string ra1, std::string dec1, 
		std::string ra2, std::string dec2)
  {
    /**
     *  Calculates the angular separation between two sky positions,
     *  given as strings for RA and DEC. Uses the function
     *  angSep(double,double,double,double).
     * \param ra1 The right ascension for the first position.
     * \param dec1 The declination for the first position.
     * \param ra2 The right ascension for the second position.
     * \param dec2 The declination for the second position.
     * \return The angular separation in degrees.
     */
    if((ra1==ra2)&&(dec1==dec2)) 
      return 0.;
    else
      return angSep(dmsToDec(ra1)*15.,dmsToDec(dec1),
		    dmsToDec(ra2)*15.,dmsToDec(dec2) );
  }

  double angSep(double ra1, double dec1, double ra2, double dec2)
  {
    /**
     *  Calculates the angular separation between two sky positions,
     *  where RA and DEC are given in decimal degrees.
     * \param ra1 The right ascension for the first position.
     * \param dec1 The declination for the first position.
     * \param ra2 The right ascension for the second position.
     * \param dec2 The declination for the second position.
     * \return The angular separation in degrees.
     */
    double r1,d1,r2,d2;
    r1 = ra1  * M_PI / 180.;
    d1 = dec1 * M_PI / 180.;
    r2 = ra2  * M_PI / 180.;
    d2 = dec2 * M_PI / 180.;
    long double angsep = cos(r1-r2)*cos(d1)*cos(d2) + sin(d1)*sin(d2);
    
    return acosl(angsep)*180./M_PI;
    
  }
  //------------------------------------------------------------------------

  void equatToGal(double ra, double dec, double &gl, double &gb)
  {
    /**
     * Converts an equatorial (ra,dec) position to galactic
     * coordinates. The equatorial position is assumed to be J2000.0.
     *
     * \param ra Right Ascension, J2000.0
     * \param dec Declination, J2000.0
     * \param gl Galactic Longitude. Returned value.
     * \param gb Galactic Latitude. Returned value.
     */

    const double NGP_RA = 192.859508 * M_PI/180.;
    const double NGP_DEC= 27.128336 * M_PI/180.;
    const double ASC_NODE=32.932;

    double deltaRA = ra * M_PI/180. - NGP_RA;
    double d = dec     * M_PI/180.;
    double sinb = cos(d) * cos(NGP_DEC) * cos(deltaRA) + sin(d) * sin(NGP_DEC);
    gb = asin(sinb);

    // The l in sinl and cosl here is really gl-ASC_NODE
    double sinl = (sin(d)*cos(NGP_DEC) - cos(d)*cos(deltaRA)*sin(NGP_DEC)) / cos(gb);
    double cosl = cos(d) * sin(deltaRA) / cos(gb);
    // atan returns a value between -90 and 90 degrees.
    // Need to correct the value of l according to the correct quandrant it is in.
    // This is worked out using the signs of sinl and cosl
    if(sinl>0){
      if(cosl>0) gl = atan(sinl/cosl);
      else       gl = atan(sinl/cosl) + M_PI;
    }
    else{
      if(cosl>0) gl = atan(sinl/cosl) + 2.*M_PI;
      else       gl = atan(sinl/cosl) + M_PI;
    }

    // Find the correct values of the lat & lon in degrees.
    gb = asin(sinb) * 180. / M_PI;
    gl = gl * 180. / M_PI + ASC_NODE;
  }
  //------------------------------------------------------------------------

  unsigned char reverseByte(unsigned char c)
  {
    /**
     *  Reverses the bits in a char (a byte), by using a 256-element
     *  look-up table. This means 01100011 becomes 11000110.
     *  /param c The byte, as a char, to be reversed.
     *  /return The reversed byte.
     */

    static const unsigned char BitReverseTable256[] = 
      {
	0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 
	0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 
	0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 
	0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 
	0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 
	0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
	0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 
	0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
	0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
	0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 
	0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
	0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
	0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 
	0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
	0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 
	0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
      };
    
    unsigned char t = c;
    unsigned char r = BitReverseTable256[c & 0xff];
    return r;

  }

  unsigned char bitToggle(unsigned char c)
  {
    /**
     *  Toggles each bit in a byte, changing 0s to 1s and 1s to 0s.
     *  We use a lookup table for transparency -- could easily just
     *  use the complement bit function ~, but this is just as good.
     * \param c Byte to toggle
     * \return Byte with bits toggled (ooh matron!)
     */
    const unsigned char BitToggleTable256[] = 
      {
	0xff, 0xfe, 0xfd, 0xfc, 0xfb, 0xfa, 0xf9, 0xf8, 0xf7, 0xf6, 0xf5, 0xf4, 0xf3, 0xf2, 0xf1, 0xf0, 
	0xef, 0xee, 0xed, 0xec, 0xeb, 0xea, 0xe9, 0xe8, 0xe7, 0xe6, 0xe5, 0xe4, 0xe3, 0xe2, 0xe1, 0xe0, 
	0xdf, 0xde, 0xdd, 0xdc, 0xdb, 0xda, 0xd9, 0xd8, 0xd7, 0xd6, 0xd5, 0xd4, 0xd3, 0xd2, 0xd1, 0xd0, 
	0xcf, 0xce, 0xcd, 0xcc, 0xcb, 0xca, 0xc9, 0xc8, 0xc7, 0xc6, 0xc5, 0xc4, 0xc3, 0xc2, 0xc1, 0xc0, 
	0xbf, 0xbe, 0xbd, 0xbc, 0xbb, 0xba, 0xb9, 0xb8, 0xb7, 0xb6, 0xb5, 0xb4, 0xb3, 0xb2, 0xb1, 0xb0, 
	0xaf, 0xae, 0xad, 0xac, 0xab, 0xaa, 0xa9, 0xa8, 0xa7, 0xa6, 0xa5, 0xa4, 0xa3, 0xa2, 0xa1, 0xa0, 
	0x9f, 0x9e, 0x9d, 0x9c, 0x9b, 0x9a, 0x99, 0x98, 0x97, 0x96, 0x95, 0x94, 0x93, 0x92, 0x91, 0x90, 
	0x8f, 0x8e, 0x8d, 0x8c, 0x8b, 0x8a, 0x89, 0x88, 0x87, 0x86, 0x85, 0x84, 0x83, 0x82, 0x81, 0x80, 
	0x7f, 0x7e, 0x7d, 0x7c, 0x7b, 0x7a, 0x79, 0x78, 0x77, 0x76, 0x75, 0x74, 0x73, 0x72, 0x71, 0x70, 
	0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x69, 0x68, 0x67, 0x66, 0x65, 0x64, 0x63, 0x62, 0x61, 0x60, 
	0x5f, 0x5e, 0x5d, 0x5c, 0x5b, 0x5a, 0x59, 0x58, 0x57, 0x56, 0x55, 0x54, 0x53, 0x52, 0x51, 0x50, 
	0x4f, 0x4e, 0x4d, 0x4c, 0x4b, 0x4a, 0x49, 0x48, 0x47, 0x46, 0x45, 0x44, 0x43, 0x42, 0x41, 0x40, 
	0x3f, 0x3e, 0x3d, 0x3c, 0x3b, 0x3a, 0x39, 0x38, 0x37, 0x36, 0x35, 0x34, 0x33, 0x32, 0x31, 0x30, 
	0x2f, 0x2e, 0x2d, 0x2c, 0x2b, 0x2a, 0x29, 0x28, 0x27, 0x26, 0x25, 0x24, 0x23, 0x22, 0x21, 0x20, 
	0x1f, 0x1e, 0x1d, 0x1c, 0x1b, 0x1a, 0x19, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 
	0x0f, 0x0e, 0x0d, 0x0c, 0x0b, 0x0a, 0x09, 0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01, 0x00
      };

    unsigned char d = BitToggleTable256[c];
    
//    unsigned char d = ~c;

    return d;

  }

}
